/***************************************************************************
                          kflamesaver.cpp  -  description
                             -------------------
    begin                : Sun Dec  8 20:31:43 ART 2002
    copyright            : (C) 2002 by Pedro J. Romero (Cronel)
    email                : pejro@millicom.com.ar
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include <qpainter.h>
#include <qtimer.h>
#include <qimage.h>
#include <qbitmap.h>

#include <klocale.h>
#include <kglobal.h>
#include <kconfig.h>
#include <kstandarddirs.h>

#include "kflamesaver.h"
#include "kflameconfig.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
#define VERSION "(unknown)"
#endif

extern "C"
{
        const char *kss_applicationName = "kflamesaver";
        const char *kss_description = I18N_NOOP( "KFlameSaver" );
        const char *kss_version = VERSION;

        KScreenSaver *kss_create( WId d )
        {
                return new KFlameSaver(d);
        }

        QDialog *kss_setup()
        {
                return new KFlameConfig(0L,0L);
        }
}


/***********************************************************************
*/

KTextInfo::KTextInfo( QString str, int frame, int x, int y, int col ) :
        m_text(str), m_x(x), m_y(y), m_frame(frame), m_col(col)
{
}

KTextInfo::KTextInfo( QString str )
{
        setFromString( str );
}

KTextInfo::~KTextInfo()
{
}

void KTextInfo::setFromString( QString str )
{
        QStringList list;
        list = QStringList::split( "||", str );
        m_frame = list[0].toInt();
        m_x = list[1].toInt();
        m_y = list[2].toInt();
        m_col = list[3].toInt();

        list.remove( list.at(0) );
        list.remove( list.at(0) );
        list.remove( list.at(0) );
        list.remove( list.at(0) );

        m_text = list.join( "||" );
}

QString KTextInfo::toString()
{
        QStringList list;
        list.append( QString("%1").arg(m_frame) );
        list.append( QString("%1").arg(m_x) );
        list.append( QString("%1").arg(m_y) );
        list.append( QString("%1").arg(m_col) );
        list.append( m_text );
        return list.join( "||" );
}


/***********************************************************************
*/

void KFlameSaver::readFromConfig()
{
        using namespace defaults;
        KConfig *cfg = KGlobal::config();
        KConfigGroupSaver save(cfg, "Settings" );
        uint i, count;
        QString str;
        KTextInfo *ti;

        m_mode = cfg->readEntry( "Mode" );

        if( !m_mode.compare("Twin Peaks") )
        {
                QString fname;
                fname = cfg->readEntry( "Pic", defaultPic );
                m_pix.load( fname );

                float ratio = (float)m_pix.height() / (float)m_pix.width();
                int w, h;
                if( ratio > 1.0 )
                {
                        h = height();
                        w = (int) (h/ratio);
                }
                else
                {
                        w = (int) (width()/3.0*2.0);
                        h = (int) (ratio * w);
                }
                QPixmap buf(w,h);
                QPainter p(&buf);
                p.drawPixmap( buf.rect(), m_pix );
                p.end();
                m_pix = buf;

                /*
                0..100                   flame
                100..117+                text
                117+..325                flame
                325..400                 frozen flame
                400..ext                 drapes open
                400+ext..400+ext+50      frozen. nice pic.
                400+ext+50..400+ext*2+50 drapes close
                400+ext*2+50...1000      flame
                */
                int cw, ch;
                cw = (width() + 239) / 240;
                ch = (height() + 159) / 160;
                setCellSize( cw, ch );

                m_drapeExt = m_width / 3;
                if( m_drapeExt * m_cellWidth > m_pix.width()/2 )
                {
                        m_drapeExt = m_pix.width() / 2 / m_cellWidth;
                }
                printf("%d %d\n", m_drapeExt * m_cellWidth, m_pix.width()/2 );

                setFlame( 2, 1, 1 );
                setFlameCol( Qt::red );
                setTextFrames( 1000 + m_drapeExt * 2 + 50);
                setFont( cfg->readFontEntry( "TextFont", &defaultTextFont ) );
                addText( "Fire", 100, CENTER_EXT, m_height-40, 255 );
                addText( "Fire", 101, CENTER_EXT, m_height-40-2, 255 );
                addText( "Fire", 102, CENTER_EXT, m_height-40-4, 255 );
                addText( "Walk With Me", 115, CENTER_EXT, m_height-40, 255 );
                addText( "Walk With Me", 116, CENTER_EXT, m_height-42, 255 );
                addText( "Walk With Me", 117, CENTER_EXT, m_height-44, 255 );
        }
        else
        {
                if( !cfg->readBoolEntry( "AutoCellSize", defaultAutoCellSize ) )
                        setCellSize( cfg->readNumEntry("CellWidth", defaultCellWidth ),
                                cfg->readNumEntry("CellHeight", defaultCellHeight) );
                else
                        setCellSize( (width() + 239) / 240, (height() + 159) / 160 );
                setFlame( cfg->readNumEntry( "HotSpots", defaultHotSpots ) ,
                        cfg->readNumEntry( "FadeThreshold", defaultFadeThreshold ),
                        cfg->readNumEntry( "FadeAmount", defaultFadeAmount ) );
                setFlameCol( cfg->readColorEntry( "FlameColor", &defaultFlameColor ) );

                setTextFrames( cfg->readNumEntry( "TextFrames", defaultTextFrames ) );
                setFont( cfg->readFontEntry( "TextFont", &defaultTextFont ) );
                count = cfg->readNumEntry( "TextCount", 0 );
                for( i=0; i<count; i++ )
                {
                        str = cfg->readEntry( QString("Text_%1").arg(i) );
                        ti = new KTextInfo( str );
                        m_text.append( ti );
                }
        }
}


void KFlameSaver::setCellSize( int cellWidth, int cellHeight )
{
        m_cellWidth = cellWidth;
        m_cellHeight = cellHeight;
        m_width = width() / m_cellWidth;
        m_height = height() / m_cellHeight + m_lowerMargin;
}


void KFlameSaver::setFlame( int hotspots, int decThreshold, int decAmount )
{
        m_hotspots = hotspots;
        m_decThreshold = decThreshold;
        m_decAmount = decAmount;
}


void KFlameSaver::setFlameCol( QColor col )
{
        m_flameCol = col;
}


void KFlameSaver::init()
{
        int i, j;

        freeBuffers();

        /* set up the buffers and the work QImage */
        m_bufWork = new int*[m_width];
        m_bufScreen = new int*[m_width];
        for(i=0; i<m_width; i++)
        {
                m_bufWork[i] = new int[m_height];
                m_bufScreen[i] = new int[m_height];

                for(j=0; j<m_height;j++)
                {
                        m_bufWork[i][j] = 0;
                        m_bufScreen[i][j] = 0;
                }
        }

        img.create( m_width, m_height-m_lowerMargin, 8, 256 );

        /* Set up the palette:
           0   ...    128 ..          255
           black     desired color    white
        */
        float fr, fg, fb, ir, ig, ib;
        ir = (float)m_flameCol.red()/128.0;
        ig = (float)m_flameCol.green()/128.0;
        ib = (float)m_flameCol.blue()/128.0;
        fr = fg = fb = 0.0;        
        for(i=0;i<128;i++)
        {
                img.setColor( i, qRgb( (int)fr, (int)fg, (int)fb ) );
                fr += ir;
                fg += ig;
                fb += ib;
        }

        ir = (float)(255-m_flameCol.red()) / 128;
        ig = (float)(255-m_flameCol.green()) / 128;
        ib = (float)(255-m_flameCol.blue()) / 128;
        
        fr = m_flameCol.red();
        fg = m_flameCol.green();
        fb = m_flameCol.blue();
        for(i=0;i<128;i++)
        {
                img.setColor( i+128, qRgb( (int)fr, (int)fg, (int)fb) );
                fr += ir;
                fg += ig;
                fb += ib;
        }

        /* paint ourselves black in case we dont cover all the screen
          (i.e. if screen w/h isnt divisible by cell w/h) */
        QPainter p(this);
        p.fillRect( rect(), Qt::black );
        p.end();

        m_curFrame = 0;
}

void KFlameSaver::setTextFrames( int frames )
{
        m_text.clear();
        m_textFrames = frames;
        m_curFrame = 0;
}

void KFlameSaver::addText( QString text, int frame, int x, int y, int col )
{
        m_text.append( new KTextInfo( text, frame, x, y, col ) );
}


KFlameSaver::KFlameSaver(WId id) : KScreenSaver( id )
{
        m_text.setAutoDelete( true );
        srandom( time(NULL) );

        m_bufWork = m_bufScreen = NULL;
        m_lowerMargin = 3;
        readFromConfig();
        init();

        timer = new QTimer(this);
        if( !m_mode.compare("Twin Peaks") )
        {
                connect( timer, SIGNAL(timeout()), this, SLOT(twinPeaks()) );
                twinPeaks();
        }
        else
        {
                connect( timer, SIGNAL(timeout()), this, SLOT(flame()) );
                flame();
        }
        timer->start( 0 );
}


void KFlameSaver::freeBuffers()
{
        int i;

        if( !m_bufScreen )
                return;
        for(i=0; i<m_width; i++ )
        {
                delete [] m_bufScreen[i];
                delete [] m_bufWork[i];
        }
        delete [] m_bufScreen;
        delete [] m_bufWork;
}


KFlameSaver::~KFlameSaver()
{
        freeBuffers();
}



void KFlameSaver::randomHotSpots( int chance )
{
        int i;
        for(i=0;i<m_width;i++)
        {
                m_bufScreen[i][m_height-1] = 0;
                if( (random()%chance) == 0 )
                {
                        m_bufScreen[i][m_height-1] = 255;
                }
        }
}


int KFlameSaver::avgCol( int x, int y )
{
        int c = 0;
        int *p;
        
        p = m_bufScreen[x-1];
        c += p[y-1];
        c += p[y];
        c += p[y+1];

        p = m_bufScreen[x];
        c += p[y-1];
        c += p[y];
        c += p[y+1];

        p = m_bufScreen[x+1];
        c += p[y-1];
        c += p[y];
        c += p[y+1];
        return c/9;
}

int KFlameSaver::safeAvgCol( int x, int y )
{
        int c, t;

        c = 1;
        t = m_bufScreen[x][y];

        if( x < m_width-1 )
        {
                c++;
                t += m_bufScreen[x+1][y];
        }

        if( x > 0 )
        {
                c++;
                t += m_bufScreen[x-1][y];
        }

        if( y < m_height-1 )
        {
                c++;
                t += m_bufScreen[x][y+1];
                if( x > 0 )
                {
                        c++;
                        t += m_bufScreen[x-1][y+1];
                }
                if( x < m_width-1 )
                {
                        c++;
                        t += m_bufScreen[x+1][y+1];
                }
        }
        if( y > 0 )
        {
                c++;
                t += m_bufScreen[x][y-1];
                if( x > 0 )
                {
                        c++;
                        t += m_bufScreen[x-1][y-1];
                }
                if( x < m_width-1 )
                {
                        c++;
                        t += m_bufScreen[x+1][y-1];
                }
        }
        return t/c;
}


void KFlameSaver::handleText( int **buf )
{
        int i;
        KTextInfo *ti;
        for( i=0; (uint)i<m_text.count(); i++ )
        {
                ti = m_text.at(i);
                if( ti->m_frame == m_curFrame )
                        putString( ti->m_x, ti->m_y, ti->m_text, buf, ti->m_col );
        }
}

void KFlameSaver::putString( int dx, int dy, QString str, int **buf, int col )
{
        QFontMetrics fm(font());
        int w, h;

        w = fm.width( str );
        h = fm.height();

        QBitmap bm( w, h );
        bm.fill( Qt::color0 );
        QPainter p;
        p.begin( &bm );
        p.setPen( Qt::color1 );
        p.setFont( font() );
        p.drawText( 0, h, str );
        p.end();

        QImage img = bm.convertToImage();
        img = img.convertDepth( 8 );

        if( dx == CENTER_EXT )
                dx = (m_width - img.width())/2;
        else if( dx < 0 )
        {
                dx = -dx;
                dx -= img.width()/2;
        }
        if( dy == CENTER_EXT )
                dy = (m_height - img.height()) / 2;
        if( dy < 0 )
        {
                dy = -dy;
                dy -= img.height()/2;
        }

        if( dx < 0 || dy < 0 )
                return;

        int y, x, maxw;
        if( img.width()+dx > m_width )
                maxw = m_width - dx;
        else
                maxw = img.width();
        for(y=0; y<img.height(); y++ )
        {
                uchar *line = img.scanLine( y );
                for( x=0; x<maxw; x++ )
                {
                        if( line[x] != 0 )
                        {
                                buf[x+dx][y+dy] = col;
                        }
                }
        }
}


void KFlameSaver::renderBuffer( QPainter *p, int **buf )
{
        QPixmap pix;
        int i, j;

        for(i=0;i<m_height-m_lowerMargin;i++)
        {
                uchar *line = img.scanLine(i);
                for(j=0;j<m_width;j++)
                        line[j] = buf[j][i];
        }
        pix.convertFromImage( img );
        p->scale( m_cellWidth, m_cellHeight );
        p->drawPixmap( 0, 0, pix );
}

void KFlameSaver::averageBuffer()
{
        int i, j, c;
        /* Notice the double scroll up: once when averaging, once when
           copying the buffers. Perhaps not wise but it looks the best */
        for(i=1;i<m_width-1;i++)
        {
                for(j=1;j<m_height-2;j++)
                {
                        c = avgCol( i, j+1 );
                        if( c > m_decThreshold )
                                c -= m_decAmount;
                        else
                                c = 0;
                       m_bufWork[i][j] =  c;
                }
        }
        for(i=0;i<m_height-1;i++)
        {
                m_bufWork[0][i] = safeAvgCol( 0, i+1 );
                m_bufWork[m_width-1][i] = safeAvgCol( m_width-1, i+1 );
        }
        for(i=0;i<m_width;i++)
        {
                m_bufWork[i][0] = safeAvgCol( i, 0+1 );
                m_bufWork[i][m_height-2] = safeAvgCol( i, m_height-1 );
                m_bufWork[i][m_height-1] = m_bufScreen[i][m_height-1];
        }
        for(i=0;i<m_width;i++)
        {
                for(j=1;j<m_height;j++)
                        m_bufScreen[i][j-1] = m_bufWork[i][j];
        }
}

void KFlameSaver::flame()
{
        QPixmap pix;

        randomHotSpots( m_hotspots );
        averageBuffer();
        if( m_textFrames )
        {
                handleText( m_bufScreen );
                m_curFrame++;
                m_curFrame %= m_textFrames;
        }
        QPainter p(this);
        renderBuffer( &p, m_bufWork );
        p.end();
}

void KFlameSaver::drapes()
{
        int i, j;
        QPixmap pix;

        for(i=0; i<m_width; i++ )
        {
                for(j=0; j<m_height; j++ )
                        m_bufScreen[i][j] = 0;
        }
        for(i=0;i<=m_width/2 - m_drapeOff; i++)
        {
                for(j=0;j<m_height;j++)
                {
                        m_bufScreen[i][j] = m_bufWork[i+m_drapeOff][j];
                        m_bufScreen[m_width-i-1][j] = m_bufWork[m_width-i-m_drapeOff-1][j];
                }
        }

        QPixmap buf( size() );
        
        QPainter p(&buf);
        renderBuffer( &p, m_bufScreen );
        p.end();

        if( m_drapeOff > 1 )
        {
                p.begin(&buf);
                int doff = (m_drapeOff-1) * m_cellWidth;
                p.drawPixmap(
                        width()/2 - doff, (height()-m_pix.height())/2,
                        m_pix,                
                        m_pix.width()/2 - doff, 0,
                        doff*2+1, m_pix.height()
                        );
                p.end();
        }

        p.begin( this );
        p.drawPixmap( 0, 0, buf );
        p.end();
}

void KFlameSaver::twinPeaks()
{
        if( m_curFrame < 400 )
        {
                if( m_curFrame == 0 )
                        setFlame( 2, 1, 1 );
                if( m_curFrame < 300 )
                {
                        randomHotSpots(m_hotspots);
                }
                else if( m_curFrame == 300 )
                {
                        int i;
                        for( i=0; i<=6; i++ )
                                m_bufScreen[m_width/2-3+i][m_height-1] = 0;
                        setFlame( 2, 0, 0 );
                }
                averageBuffer();
                handleText( m_bufScreen );
                QPainter p(this);
                renderBuffer( &p, m_bufWork );
                p.end();
        }
        else if( m_curFrame < 400 + m_drapeExt * 2 + 50 )
        {
                if( m_curFrame == 400 )
                        m_drapeOff = 0;
                int dframe = m_curFrame - 400;
                if( dframe < m_drapeExt )
                        m_drapeOff++;
                else if( dframe >= m_drapeExt + 50 )
                        m_drapeOff--;
                drapes();
        }
        else
        {
                if( m_curFrame == 400 + m_drapeExt * 2 + 50 )
                        setFlame( 2, 1, 1 );
                randomHotSpots( m_hotspots );
                averageBuffer();
                QPainter p(this);
                renderBuffer( &p, m_bufWork );
                p.end();
        }
        m_curFrame++;
        m_curFrame %= m_textFrames;
}

void KFlameSaver::keyPressEvent( QKeyEvent *ke )
{
        switch(ke->key())
        {
        case Key_Escape:
                timer->stop();
                break;
        case Key_F1:
                timer->start( 0 );
                break;
        case Key_Space:
                flame();
                update();
                break;
        case Key_T:
                twinPeaks();
                update();
                break;
        case Key_C:
                KFlameConfig *dlg = new KFlameConfig( this, 0 );
                dlg->show();
                break;
        }

}
