/***************************************************************************
                          KFlameSaver.h  -  description
                             -------------------
    begin                : Sun Dec  8 20:31:43 ART 2002
    copyright            : (C) 2002 by Pedro J. Romero (Cronel)
    email                : pejro@millicom.com.ar
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KFlameSaver_H
#define KFlameSaver_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <qwidget.h>
#include <qimage.h>

#include <kapp.h>
#include <kscreensaver.h>

class KFlameSaver;

/** Simple class to hold info on text to be displayed */
class KTextInfo
{
    public:
        QString         m_text;
        int             m_x, m_y, m_frame, m_col;
        KTextInfo(QString str, int frame, int x, int y, int col );
        KTextInfo( QString config_str );
        ~KTextInfo();

        QString toString();
        void setFromString( QString str );
};

#define CENTER_EXT (-1)

/** KFlameSaver is the base class of the project */
class KFlameSaver : public KScreenSaver
{
    Q_OBJECT
    public:
        /** constructor */
        KFlameSaver( WId id=0 );
        /** destructor */
        ~KFlameSaver();

        /** Set cell size for screensaver */
        void setCellSize( int cellWidth, int cellHeight );
        /** Set flame parameters */
        void setFlame( int hotspots, int decThreshold, int decAmount );
        /** Set flame color */
        void setFlameCol( QColor col );

        /** Set number of text frames */
        void setTextFrames( int frames );
        /** Add text in a frame */
        void addText( QString text, int frame, int x, int y, int col );

        /** Init yourself after having set everything */
        void init();

    protected:
        /** For testing */
        void keyPressEvent( QKeyEvent *ke );
    protected slots:
        void flame();
        void drapes();
        void twinPeaks();
    private:
        int             m_width, m_height;
        QPixmap         m_pix;
        QImage          img;
        int             **m_bufScreen;
        int             **m_bufWork;
        QTimer          *timer;
        QString         m_mode;

        int             m_lowerMargin;
        
        /* flame config */
        int             m_cellWidth, m_cellHeight;
        int             m_hotspots;
        int             m_decThreshold, m_decAmount;
        QColor          m_flameCol;

        /* text config */
        int                     m_textFrames, m_curFrame;
        QList<KTextInfo>        m_text;

        int             m_drapeOff, m_drapeExt;

        void handleText( int **buf );
        /** Draws a string on the buffer buf, at dx,dy in color index col */
        void putString( int dx, int dy, QString str, int **buf, int col );
        /** Return average; assumes its not at the border */
        int avgCol( int x, int y );
        /** Return average */
        int safeAvgCol( int x, int y );
        void freeBuffers();
        /** Add random hot spots to the last row of screen buffer */
        void randomHotSpots( int chance );
        /** Self configure from config file */
        void readFromConfig();
        /** Render buffer buf using painter p */
        void renderBuffer( QPainter *p, int **buf );
        /** Do averaging on screen buffer */
        void averageBuffer();
};

#endif
